// GenericDriver.cpp
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2011 Mike McCauley
// Contributed by Joanna Rutkowska
// $Id: GenericSPI.h,v 1.8 2017/11/06 00:04:08 mikem Exp $
// Modified by IndustrialShields

#include <GenericDriver.h>

GenericDriver::GenericDriver()
    :
    _mode(RHModeInitialising),
    _thisAddress(RH_BROADCAST_ADDRESS),
    _txHeaderTo(RH_BROADCAST_ADDRESS),
    _txHeaderFrom(RH_BROADCAST_ADDRESS),
    _txHeaderId(0),
    _txHeaderFlags(0),
    _rxBad(0),
    _rxGood(0),
    _txGood(0),
    _cad_timeout(0)
{
}

bool GenericDriver::init()
{
    return true;
}

// Blocks until a valid message is received
void GenericDriver::waitAvailable()
{
    while (!available())
	YIELD;
}

// Blocks until a valid message is received or timeout expires
// Return true if there is a message available
// Works correctly even on millis() rollover
bool GenericDriver::waitAvailableTimeout(uint16_t timeout)
{
    unsigned long starttime = millis();
    while ((millis() - starttime) < timeout)
    {
        if (available())
	{
           return true;
	}
	YIELD;
    }
    return false;
}

bool GenericDriver::waitPacketSent()
{
    while (_mode == RHModeTx)
	YIELD; // Wait for any previous transmit to finish
    return true;
}

bool GenericDriver::waitPacketSent(uint16_t timeout)
{
    unsigned long starttime = millis();
    while ((millis() - starttime) < timeout)
    {
        if (_mode != RHModeTx) // Any previous transmit finished?
           return true;
	YIELD;
    }
    return false;
}

// Wait until no channel activity detected or timeout
bool GenericDriver::waitCAD()
{
    if (!_cad_timeout)
	return true;

    // Wait for any channel activity to finish or timeout
    // Sophisticated DCF function...
    // DCF : BackoffTime = random() x aSlotTime
    // 100 - 1000 ms
    // 10 sec timeout
    unsigned long t = millis();
    while (isChannelActive())
    {
         if (millis() - t > _cad_timeout) 
	     return false;
#if (RH_PLATFORM == RH_PLATFORM_STM32) // stdlib on STMF103 gets confused if random is redefined
	 delay(_random(1, 10) * 100);
#else
         delay(random(1, 10) * 100); // Should these values be configurable? Macros?
#endif
    }

    return true;
}

// subclasses are expected to override if CAD is available for that radio
bool GenericDriver::isChannelActive()
{
    return false;
}

void GenericDriver::setPromiscuous(bool promiscuous)
{
    _promiscuous = promiscuous;
}

void GenericDriver::setThisAddress(uint8_t address)
{
    _thisAddress = address;
}

void GenericDriver::setHeaderTo(uint8_t to)
{
    _txHeaderTo = to;
}

void GenericDriver::setHeaderFrom(uint8_t from)
{
    _txHeaderFrom = from;
}

void GenericDriver::setHeaderId(uint8_t id)
{
    _txHeaderId = id;
}

void GenericDriver::setHeaderFlags(uint8_t set, uint8_t clear)
{
    _txHeaderFlags &= ~clear;
    _txHeaderFlags |= set;
}

uint8_t GenericDriver::headerTo()
{
    return _rxHeaderTo;
}

uint8_t GenericDriver::headerFrom()
{
    return _rxHeaderFrom;
}

uint8_t GenericDriver::headerId()
{
    return _rxHeaderId;
}

uint8_t GenericDriver::headerFlags()
{
    return _rxHeaderFlags;
}

int16_t GenericDriver::lastRssi()
{
    return _lastRssi;
}

GenericDriver::RHMode  GenericDriver::mode()
{
    return _mode;
}

void  GenericDriver::setMode(RHMode mode)
{
    _mode = mode;
}

bool  GenericDriver::sleep()
{
    return false;
}

// Diagnostic help
void GenericDriver::printBuffer(const char* prompt, const uint8_t* buf, uint8_t len)
{
#ifdef RH_HAVE_SERIAL
    Serial.println(prompt);
    uint8_t i;
    for (i = 0; i < len; i++)
    {
	if (i % 16 == 15)
	    Serial.println(buf[i], HEX);
	else
	{
	    Serial.print(buf[i], HEX);
	    Serial.print(' ');
	}
    }
    Serial.println("");
#endif
}

uint16_t GenericDriver::rxBad()
{
    return _rxBad;
}

uint16_t GenericDriver::rxGood()
{
    return _rxGood;
}

uint16_t GenericDriver::txGood()
{
    return _txGood;
}

void GenericDriver::setCADTimeout(unsigned long cad_timeout)
{
    _cad_timeout = cad_timeout;
}

#if (RH_PLATFORM == RH_PLATFORM_ARDUINO) && defined(RH_PLATFORM_ATTINY)
// Tinycore does not have __cxa_pure_virtual, so without this we
// get linking complaints from the default code generated for pure virtual functions
extern "C" void __cxa_pure_virtual()
{
    while (1);
}
#endif
