// LoRa.cpp
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2011 Mike McCauley
// Contributed by Joanna Rutkowska
// $Id: GenericSPI.h,v 1.8 2017/11/06 00:04:08 mikem Exp $
// Modified by IndustrialShields

#include <LoRa.h>

#ifdef HAVE_RFM9X

// interrupt handler and related code must be in RAM on ESP8266,
// according to issue #46.
#if defined(ESP8266)
    #define INTERRUPT_ATTR ICACHE_RAM_ATTR
#else
    #define INTERRUPT_ATTR
#endif

// #define SERIAL_DEBUG // Uncomment to recieve debug information over serial

// Interrupt vectors for the 3 Arduino interrupt pins
// Each interrupt can be handled by a different instance of RF95, allowing you to have
// 2 or more LORAs per Arduino
LoRa* LoRa::_deviceForInterrupt[RF95_NUM_INTERRUPTS] = {0, 0, 0};
uint8_t LoRa::_interruptCount = 0; // Index into _deviceForInterrupt for next device

// These are indexed by the values of ModemConfigChoice
// Stored in flash (program) memory to save SRAM
PROGMEM static const LoRa::ModemConfig MODEM_CONFIG_TABLE[] =
{
    //  1d,     1e,      26
    { 0x72,   0x74,    0x04}, // Bw125Cr45Sf128 (the chip default), AGC enabled
    { 0x92,   0x74,    0x04}, // Bw500Cr45Sf128, AGC enabled
    { 0x48,   0x94,    0x04}, // Bw31_25Cr48Sf512, AGC enabled
    { 0x78,   0xc4,    0x0c}, // Bw125Cr48Sf4096, AGC enabled

};

//uint8_t slaveSelectPin, uint8_t interruptPin, GenericSPI& spi
LoRa::LoRa()
    :
    SPIDriver(RFM9X_RESET_PIN, hardware_spi),
    _rxBufValid(0)
{
    _slaveSelectPin = RFM9X_SS_PIN;
    _interruptPin = RFM9X_INT_PIN;
    _myInterruptIndex = 0xff; // Not allocated ye
}

bool LoRa::begin()
{
    if (!SPIDriver::begin()){
      #ifdef SERIAL_DEBUG
        Serial.println(F("ERROR: Failed to initialize SPI Driver."));
      #endif
      return false;
    }


    // Determine the interrupt number that corresponds to the interruptPin
    int interruptNumber = digitalPinToInterrupt(_interruptPin);
    if (interruptNumber == NOT_AN_INTERRUPT){
      #ifdef SERIAL_DEBUG
        Serial.println(F("ERROR: Provided pin does not support interrupts."));
      #endif
      return false;
    }

#ifdef ATTACHINTERRUPT_TAKES_PIN_NUMBER
    interruptNumber = _interruptPin;
#endif

    // Tell the low level SPI interface we will use SPI within this interrupt
    spiUsingInterrupt(interruptNumber);

    // No way to check the device type :-(

    // Set sleep mode, so we can also set LORA mode:
    spiWrite(RF95_REG_01_OP_MODE, RF95_MODE_SLEEP | RF95_LONG_RANGE_MODE);
    delay(10); // Wait for sleep mode to take over from say, CAD
    // Check we are in sleep mode, with LORA set
    if (spiRead(RF95_REG_01_OP_MODE) != (RF95_MODE_SLEEP | RF95_LONG_RANGE_MODE))
    {
      #ifdef SERIAL_DEBUG
        Serial.println(F("ERROR: Failed to put device in LoRa mode."));
        Serial.print(F("RF95_REG_01_OP_MODE: "));
        Serial.println(spiRead(RF95_REG_01_OP_MODE), HEX);
      #endif
	    return false; // No device present?
    }

    // Add by Adrien van den Bossche <vandenbo@univ-tlse2.fr> for Teensy
    // ARM M4 requires the below. else pin interrupt doesn't work properly.
    // On all other platforms, its innocuous, belt and braces
    pinMode(_interruptPin, INPUT);

    // Set up interrupt handler
    // Since there are a limited number of interrupt glue functions isr*() available,
    // we can only support a limited number of devices simultaneously
    // ON some devices, notably most Arduinos, the interrupt pin passed in is actuallt the
    // interrupt number. You have to figure out the interruptnumber-to-interruptpin mapping
    // yourself based on knwledge of what Arduino board you are running on.
    if (_myInterruptIndex == 0xff)
    {
	     // First run, no interrupt allocated yet
	     if (_interruptCount <= RF95_NUM_INTERRUPTS)
	      _myInterruptIndex = _interruptCount++;
	     else{
         #ifdef SERIAL_DEBUG
          Serial.println(F("ERROR: Insufficient interrupt vectors."));
         #endif
         return false; // Too many devices, not enough interrupt vectors
       }
    }
    _deviceForInterrupt[_myInterruptIndex] = this;
    if (_myInterruptIndex == 0)
	attachInterrupt(interruptNumber, isr0, RISING);
    else if (_myInterruptIndex == 1)
	attachInterrupt(interruptNumber, isr1, RISING);
    else if (_myInterruptIndex == 2)
	attachInterrupt(interruptNumber, isr2, RISING);
    else{
      #ifdef SERIAL_DEBUG
        Serial.println(F("ERROR: Failed to attach interrupt pin. Insufficient interrupt vectors."));
      #endif
      return false; // Too many devices, not enough interrupt vectors
    }


    // Set up FIFO
    // We configure so that we can use the entire 256 byte FIFO for either receive
    // or transmit, but not both at the same time
    spiWrite(RF95_REG_0E_FIFO_TX_BASE_ADDR, 0);
    spiWrite(RF95_REG_0F_FIFO_RX_BASE_ADDR, 0);

    // Packet format is preamble + explicit-header + payload + crc
    // Explicit Header Mode
    // payload is TO + FROM + ID + FLAGS + message data
    // RX mode is implmented with RXCONTINUOUS
    // max message data length is 255 - 4 = 251 octets

    setModeIdle();

    // Set up default configuration
    // No Sync Words in LORA mode.
    setModemConfig(Bw125Cr45Sf128); // Radio default
//    setModemConfig(Bw125Cr48Sf4096); // slow and reliable?
    setPreambleLength(8); // Default is 8
    // An innocuous ISM frequency, same as RF22's
    setFrequency(868.0);
    // Lowish power
    setTxPower(13);

    return true;
}

// C++ level interrupt handler for this instance
// LORA is unusual in that it has several interrupt lines, and not a single, combined one.
// On MiniWirelessLoRa, only one of the several interrupt lines (DI0) from the RFM95 is usefuly
// connnected to the processor.
// We use this to get RxDone and TxDone interrupts
void LoRa::handleInterrupt()
{
    // Read the interrupt register
    uint8_t irq_flags = spiRead(RF95_REG_12_IRQ_FLAGS);
    // Read the RegHopChannel register to check if CRC presence is signalled
    // in the header. If not it might be a stray (noise) packet.*
    uint8_t crc_present = spiRead(RF95_REG_1C_HOP_CHANNEL);

    if (_mode == RHModeRx
	&& ((irq_flags & (RF95_RX_TIMEOUT | RF95_PAYLOAD_CRC_ERROR))
	    | !(crc_present & RF95_RX_PAYLOAD_CRC_IS_ON)))
//    if (_mode == RHModeRx && irq_flags & (RF95_RX_TIMEOUT | RF95_PAYLOAD_CRC_ERROR))
    {
	_rxBad++;
    }
    else if (_mode == RHModeRx && irq_flags & RF95_RX_DONE)
    {
	// Have received a packet
	uint8_t len = spiRead(RF95_REG_13_RX_NB_BYTES);

	// Reset the fifo read ptr to the beginning of the packet
	spiWrite(RF95_REG_0D_FIFO_ADDR_PTR, spiRead(RF95_REG_10_FIFO_RX_CURRENT_ADDR));
	spiBurstRead(RF95_REG_00_FIFO, _buf, len);
	_bufLen = len;
	spiWrite(RF95_REG_12_IRQ_FLAGS, 0xff); // Clear all IRQ flags

	// Remember the last signal to noise ratio, LORA mode
	// Per page 111, SX1276/77/78/79 datasheet
	_lastSNR = (int8_t)spiRead(RF95_REG_19_PKT_SNR_VALUE) / 4;
 
	// Remember the RSSI of this packet, LORA mode
	// this is according to the doc, but is it really correct?
	// weakest receiveable signals are reported RSSI at about -66
	_lastRssi = spiRead(RF95_REG_1A_PKT_RSSI_VALUE);
	// Adjust the RSSI, datasheet page 87
	if (_lastSNR < 0)
	    _lastRssi = _lastRssi + _lastSNR;
	else
	    _lastRssi = (int)_lastRssi * 16 / 15;
	if (_usingHFport)
	    _lastRssi -= 157;
	else
	    _lastRssi -= 164;

	// We have received a message.
	validateRxBuf();
	if (_rxBufValid)
	    setModeIdle(); // Got one
    }
    else if (_mode == RHModeTx && irq_flags & RF95_TX_DONE)
    {
	_txGood++;
	setModeIdle();
    }
    else if (_mode == RHModeCad && irq_flags & RF95_CAD_DONE)
    {
        _cad = irq_flags & RF95_CAD_DETECTED;
        setModeIdle();
    }
    // Sigh: on some processors, for some unknown reason, doing this only once does not actually
    // clear the radio's interrupt flag. So we do it twice. Why?
    spiWrite(RF95_REG_12_IRQ_FLAGS, 0xff); // Clear all IRQ flags
    spiWrite(RF95_REG_12_IRQ_FLAGS, 0xff); // Clear all IRQ flags
}

// These are low level functions that call the interrupt handler for the correct
// instance of RF95.
// 3 interrupts allows us to have 3 different devices
void INTERRUPT_ATTR LoRa::isr0()
{
    if (_deviceForInterrupt[0])
	_deviceForInterrupt[0]->handleInterrupt();
}
void INTERRUPT_ATTR LoRa::isr1()
{
    if (_deviceForInterrupt[1])
	_deviceForInterrupt[1]->handleInterrupt();
}
void INTERRUPT_ATTR LoRa::isr2()
{
    if (_deviceForInterrupt[2])
	_deviceForInterrupt[2]->handleInterrupt();
}

// Check whether the latest received message is complete and uncorrupted
void LoRa::validateRxBuf()
{
    if (_bufLen < 4)
	return; // Too short to be a real message
    // Extract the 4 headers
    _rxHeaderTo    = _buf[0];
    _rxHeaderFrom  = _buf[1];
    _rxHeaderId    = _buf[2];
    _rxHeaderFlags = _buf[3];
    if (_promiscuous ||
	_rxHeaderTo == _thisAddress ||
	_rxHeaderTo == RH_BROADCAST_ADDRESS)
    {
	_rxGood++;
	_rxBufValid = true;
    }
}

bool LoRa::available()
{
    if (_mode == RHModeTx)
	return false;
    setModeRx();
    return _rxBufValid; // Will be set by the interrupt handler when a good message is received
}

void LoRa::clearRxBuf()
{
    ATOMIC_BLOCK_START;
    _rxBufValid = false;
    _bufLen = 0;
    ATOMIC_BLOCK_END;
}

bool LoRa::recv(uint8_t* buf, uint8_t* len)
{
    if (!available())
	return false;
    if (buf && len)
    {
	ATOMIC_BLOCK_START;
	// Skip the 4 headers that are at the beginning of the rxBuf
	if (*len > _bufLen-RF95_HEADER_LEN)
	    *len = _bufLen-RF95_HEADER_LEN;
	memcpy(buf, _buf+RF95_HEADER_LEN, *len);
	ATOMIC_BLOCK_END;
    }
    clearRxBuf(); // This message accepted and cleared
    return true;
}

bool LoRa::send(const uint8_t* data, uint8_t len)
{
    if (len > RF95_MAX_MESSAGE_LEN)
	return false;

    waitPacketSent(); // Make sure we dont interrupt an outgoing message
    setModeIdle();

    if (!waitCAD())
	return false;  // Check channel activity

    // Position at the beginning of the FIFO
    spiWrite(RF95_REG_0D_FIFO_ADDR_PTR, 0);
    // The headers
    spiWrite(RF95_REG_00_FIFO, _txHeaderTo);
    spiWrite(RF95_REG_00_FIFO, _txHeaderFrom);
    spiWrite(RF95_REG_00_FIFO, _txHeaderId);
    spiWrite(RF95_REG_00_FIFO, _txHeaderFlags);
    // The message data
    spiBurstWrite(RF95_REG_00_FIFO, data, len);
    spiWrite(RF95_REG_22_PAYLOAD_LENGTH, len + RF95_HEADER_LEN);

    setModeTx(); // Start the transmitter
    // when Tx is done, interruptHandler will fire and radio mode will return to STANDBY
    return true;
}

bool LoRa::printRegisters()
{
#ifdef HAVE_SERIAL
    uint8_t registers[] = { 0x01, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x014, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27};

    uint8_t i;
    for (i = 0; i < sizeof(registers); i++)
    {
	Serial.print(registers[i], HEX);
	Serial.print(": ");
	Serial.println(spiRead(registers[i]), HEX);
    }
#endif
    return true;
}

uint8_t LoRa::maxMessageLength()
{
    return RF95_MAX_MESSAGE_LEN;
}

bool LoRa::setFrequency(float centre)
{
    // Frf = FRF / FSTEP
    uint32_t frf = (centre * 1000000.0) / RF95_FSTEP;
    spiWrite(RF95_REG_06_FRF_MSB, (frf >> 16) & 0xff);
    spiWrite(RF95_REG_07_FRF_MID, (frf >> 8) & 0xff);
    spiWrite(RF95_REG_08_FRF_LSB, frf & 0xff);
    _usingHFport = (centre >= 779.0);

    return true;
}

void LoRa::setModeIdle()
{
    if (_mode != RHModeIdle)
    {
	spiWrite(RF95_REG_01_OP_MODE, RF95_MODE_STDBY);
	_mode = RHModeIdle;
    }
}

bool LoRa::sleep()
{
    if (_mode != RHModeSleep)
    {
	spiWrite(RF95_REG_01_OP_MODE, RF95_MODE_SLEEP);
	_mode = RHModeSleep;
    }
    return true;
}

void LoRa::setModeRx()
{
    if (_mode != RHModeRx)
    {
	spiWrite(RF95_REG_01_OP_MODE, RF95_MODE_RXCONTINUOUS);
	spiWrite(RF95_REG_40_DIO_MAPPING1, 0x00); // Interrupt on RxDone
	_mode = RHModeRx;
    }
}

void LoRa::setModeTx()
{
    if (_mode != RHModeTx)
    {
	spiWrite(RF95_REG_01_OP_MODE, RF95_MODE_TX);
	spiWrite(RF95_REG_40_DIO_MAPPING1, 0x40); // Interrupt on TxDone
	_mode = RHModeTx;
    }
}

void LoRa::setTxPower(int8_t power, bool useRFO)
{
    // Sigh, different behaviours depending on whther the module use PA_BOOST or the RFO pin
    // for the transmitter output
    if (useRFO)
    {
	if (power > 14)
	    power = 14;
	if (power < -1)
	    power = -1;
	spiWrite(RF95_REG_09_PA_CONFIG, RF95_MAX_POWER | (power + 1));
    }
    else
    {
	if (power > 23)
	    power = 23;
	if (power < 5)
	    power = 5;

	// For RF95_PA_DAC_ENABLE, manual says '+20dBm on PA_BOOST when OutputPower=0xf'
	// RF95_PA_DAC_ENABLE actually adds about 3dBm to all power levels. We will us it
	// for 21, 22 and 23dBm
	if (power > 20)
	{
	    spiWrite(RF95_REG_4D_PA_DAC, RF95_PA_DAC_ENABLE);
	    power -= 3;
	}
	else
	{
	    spiWrite(RF95_REG_4D_PA_DAC, RF95_PA_DAC_DISABLE);
	}

	// RFM95/96/97/98 does not have RFO pins connected to anything. Only PA_BOOST
	// pin is connected, so must use PA_BOOST
	// Pout = 2 + OutputPower.
	// The documentation is pretty confusing on this topic: PaSelect says the max power is 20dBm,
	// but OutputPower claims it would be 17dBm.
	// My measurements show 20dBm is correct
	spiWrite(RF95_REG_09_PA_CONFIG, RF95_PA_SELECT | (power-5));
    }
}

// Sets registers from a canned modem configuration structure
void LoRa::setModemRegisters(const ModemConfig* config)
{
    spiWrite(RF95_REG_1D_MODEM_CONFIG1,       config->reg_1d);
    spiWrite(RF95_REG_1E_MODEM_CONFIG2,       config->reg_1e);
    spiWrite(RF95_REG_26_MODEM_CONFIG3,       config->reg_26);
}

// Set one of the canned FSK Modem configs
// Returns true if its a valid choice
bool LoRa::setModemConfig(ModemConfigChoice index)
{
    if (index > (signed int)(sizeof(MODEM_CONFIG_TABLE) / sizeof(ModemConfig)))
        return false;

    ModemConfig cfg;
    memcpy_P(&cfg, &MODEM_CONFIG_TABLE[index], sizeof(LoRa::ModemConfig));
    setModemRegisters(&cfg);

    return true;
}

void LoRa::setPreambleLength(uint16_t bytes)
{
    spiWrite(RF95_REG_20_PREAMBLE_MSB, bytes >> 8);
    spiWrite(RF95_REG_21_PREAMBLE_LSB, bytes & 0xff);
}

bool LoRa::isChannelActive()
{
    // Set mode ModeCad
    if (_mode != RHModeCad)
    {
        spiWrite(RF95_REG_01_OP_MODE, RF95_MODE_CAD);
        spiWrite(RF95_REG_40_DIO_MAPPING1, 0x80); // Interrupt on CadDone
        _mode = RHModeCad;
    }

    while (_mode == RHModeCad)
        YIELD;

    return _cad;
}

void LoRa::enableTCXO()
{
    while ((spiRead(RF95_REG_4B_TCXO) & RF95_TCXO_TCXO_INPUT_ON) != RF95_TCXO_TCXO_INPUT_ON)
    {
	sleep();
	spiWrite(RF95_REG_4B_TCXO, (spiRead(RF95_REG_4B_TCXO) | RF95_TCXO_TCXO_INPUT_ON));
    }
}

// From section 4.1.5 of SX1276/77/78/79
// Ferror = FreqError * 2**24 * BW / Fxtal / 500
int LoRa::frequencyError()
{
    int32_t freqerror = 0;

    // Convert 2.5 bytes (5 nibbles, 20 bits) to 32 bit signed int
    // Caution: some C compilers make errors with eg:
    // freqerror = spiRead(RF95_REG_28_FEI_MSB) << 16
    // so we go more carefully.
    freqerror = spiRead(RF95_REG_28_FEI_MSB);
    freqerror <<= 8;
    freqerror |= spiRead(RF95_REG_29_FEI_MID);
    freqerror <<= 8;
    freqerror |= spiRead(RF95_REG_2A_FEI_LSB);
    // Sign extension into top 3 nibbles
    if (freqerror & 0x80000)
	freqerror |= 0xfff00000;

    int error = 0; // In hertz
    float bw_tab[] = {7.8, 10.4, 15.6, 20.8, 31.25, 41.7, 62.5, 125, 250, 500};
    uint8_t bwindex = spiRead(RF95_REG_1D_MODEM_CONFIG1) >> 4;
    if (bwindex < (sizeof(bw_tab) / sizeof(float)))
	error = (float)freqerror * bw_tab[bwindex] * ((float)(1L << 24) / (float)RF95_FXOSC / 500.0);
    // else not defined

    return error;
}

int LoRa::lastSNR()
{
    return _lastSNR;
}

#endif
